//Abstract base class for finite automata represented explicitly
#ifndef ABSAUTOMATA_H_
#define ABSAUTOMATA_H_
#pragma interface
#include "BitVector.h"
#include <vector>
#include <cstdio>

//Forward decs
class Atom;
class bdd;
typedef std::vector<Atom> AtomMap;
typedef std::vector<class bdd> IntBddMap;

class AbsAutomata {
  
 public:
  /**Constructor of the class 
   * @param size The number of states in the automaton
   * @param aSize The size of the alphabet
   */
  AbsAutomata(unsigned size, unsigned aSize);
  /**The destructor*/
  virtual ~AbsAutomata(); 
  /**@return The number of states in the automaton*/
  unsigned size() const {return mySize;}
  /**@return The size of the alphabet*/
  unsigned alphabetSize() const {return myAlphabetSize;}  
   /**Get a successor state
   *@param source The source state
   *@param label The Label of the transition
   *@param num Index of the transition (for non-det automatons)
   *@return The destination state
   */
  virtual unsigned dest(unsigned source, unsigned label, unsigned index) const = 0;
  /**Add a transition 
   *@param source The source state
   *@param label The label of the transition
   *@param dest The destination state
   */
  virtual void addTransition(unsigned source, unsigned label, unsigned dest) =0;
  /**Delete a transition 
   *@param source The source state
   *@param label The label of the transition
   *@param dest The destination state
   */
  virtual void delTransition(unsigned source, unsigned label, unsigned dest) =0;
  /**@return The number of transitions from a state with a given label*/ 
  virtual unsigned numArcs(unsigned source, unsigned label) const = 0;
  /**Print the automaton to a stream
   *@param output The stream where the output is directed
   *@param labels The labels of the arc in bdd form
   *@param idmap Names of the atomic proposition for the labels
   */
  void printAut(FILE *output, const IntBddMap &labels, const AtomMap &idmap) const;
   /**Print the automaton to a stream using numbers as labels
    *@param output The stream where the output is directed
    */
  void printAut(FILE *output) const;
  
  /**Print the automaton to a stream in the Maria tool format
   *@param output The stream where the output is directed
   *@param labels The labels of the arc in bdd form
   *@param idmap Names of the atomic proposition for the labels
   */
  void mariaprint(FILE *output, const IntBddMap &labels, const AtomMap &idmap) const;
  
  /**Print the automaton as a neverclaim of Spin
   *@param output The stream where the output is directed
   *@param labels The labels of the arc in bdd form
   *@param idmap Names of the atomic proposition for the labels
   */
  void spinprint(FILE *output, const IntBddMap &labels, const AtomMap &idmap ) const;

  /**Make a given state an initial state
   *@param state
   */
  
  void makeInitial(unsigned state) { 
    assert(state<mySize);
    myInitial.assign(state, true); 
  } 
  /**Remove the initial state flag for a given state
   *@param state
   */
  void delInitial(unsigned state) {
    assert(state<mySize);
    myInitial.assign(state, false); 
  }
  /**For NFA return the initial state with the lowest index. 
   * For DFA return the initial state
   * @return An initial state
   */
  unsigned initialState() const {
    unsigned i=0;
    for ( ; i<mySize; i++) {
      if (myInitial[i]) 
	break;
    }
    return i;    
  }

  /**Make a given state a final state*/
  void makeFinal(unsigned state) {
    assert(state<mySize);
    if(!myAccepting[state])
      myAccepting.assign(state, true), myNumFinal++; 
  }
  void delFinal(unsigned state) {
    assert(state<mySize);
    if(myAccepting[state])
      myAccepting.assign(state, false), myNumFinal--;    
  }
   /**@return true if the state is a final state*/
  bool isFinal(unsigned state) const {return myAccepting[state];}
  /**@return true if the state is an initial state*/
  bool isInitial(unsigned state) const {return myInitial[state];}
  /**@return The number of final states*/
  unsigned numFinal() const {return myNumFinal;}
  
 protected:
  /**Number of states of the automaton*/
  unsigned mySize;
  /**Size of the alphabet*/
  unsigned myAlphabetSize;
  /**The number of accepting states*/
  unsigned myNumFinal;
  /*Array with the accepting states*/
  class BitVector myAccepting;
  /**Array with the initial states*/  
  class BitVector myInitial;

};
#endif //ABSAUTOMATA_H_

