#pragma implementation
#include <cstdio>
#include "Automata.h"
#include "Formula.h"
#include "TemporalBinOp.h"
#include "TemporalUnOp.h"
#include "BinOp.h"
#include "Not.h"
#include "Atom.h"
#include "Const.h"

#include <hash_set>
#include <hash_map>
//static hash<int> H;
struct fHasher
{
  size_t operator()(const Formula *f) const
  {  
    return H(f->hash());
  }
};

struct fMapEq {
  bool operator() (const Formula * f1, const Formula *f2) const {
    if (f1 == f2)
      return true;
    else
      return *f1 == *f2;
  }
};

//enable fetching of formulas with hashvalues of formulas
typedef std::hash_set<Formula *, fHasher, fMapEq> FormulaSet;

//enable fetching of bddvarnumber with formula
typedef std::hash_map<Formula *, unsigned, fHasher, fMapEq> IdMap;

/**Count the number of unique subformulas*/
static unsigned 
count(const class Formula &f) {
  FormulaSet fSet;
  unsigned count=0;
  for(class Formula::Iterator t=f.newIterator(); !t.atEnd(); ++t) {
    if (fSet.find(&(*t)) == fSet.end()) { 
      fSet.insert(&(*t));
      count++;
    }
  }
  return count;
}

/**Return a BDD with contains the bdd variables for atomic propositions
 *@param formula The formula to be processed
 *@return the ap:s as a bdd
 */
bdd
labelBDD(const class Formula &f)
{  
  FormulaSet fSet;
  class bdd ap=bdd_true();
  unsigned count=0;
  for(Formula::Iterator iter=f.newIterator(); !iter.atEnd(); ++iter) {
    if (fSet.find(&(*iter)) == fSet.end()) {
      if((*iter).getType() == Formula::fAtom) {
	ap &= bdd_ithvar(count);
      }
      fSet.insert(&(*iter));
      count+=2;
    }
  }
  return ap;
}

/**Create a bdd representing the reachable states in the automaton
 * @param initial bdd for the final state expressed in primed vars
 * @param transrel bdd for the transition relation 
 * @param vars bdd for the primed state variables
 * @param repvars bddPairs for replacing the unprimed vars with primed
 */
static bdd 
reachable_states(class bdd &initial, class bdd &transrel,
			    class bdd &vars, bddPair *repvars) 
{
  class bdd reachable = initial; // Reachable state space
  class bdd prevR; // Previously reached state space and temp
  do  {
     prevR = reachable;
     class bdd newset  = bdd_appex(reachable, transrel, bddop_and, vars);
     newset = bdd_replace(newset, repvars);
     reachable |= newset;
   }
  while(prevR != reachable);
  return reachable;
} 

/**Create the transition relation of the automaton. 
 * @param idMap A map from subformulas to bdd variable numbers
 * @return The transition relation
 */
static bdd
transitions(const IdMap &idMap)
{
  class bdd transrel=bdd_true();
  //0 now refeers to the unprimed state varible of the full formula
  //1 refeers to the primed state variable...
  //i always refeers to the current formulas unprimed state variable
  //i+1 refeers to the current formulas primed state variable
  for(IdMap::const_iterator t=idMap.begin(); t!=idMap.end(); ++t) {
    unsigned i = (*t).second;
    class Formula *f=(*t).first;
    Formula::Type kind=f->getType();
    switch(kind) {
    case Formula::fTemporalBinOp:
      {
	class TemporalBinOp &temp=static_cast<TemporalBinOp &>(*f);
	if(temp.getOp() == TemporalBinOp::Until) {
	  // f = g U h  
	  // x'f <-> (x'h | (x'g & xf))
	  //unsigned temp1=(*idMap.find((temp.getRHS()))).second;
	  //unsigned temp2=(*idMap.find((temp.getLHS()))).second;
	  transrel &= bdd_biimp(bdd_ithvar(i), bdd_ithvar((*idMap.find((temp.getRHS()))).second) | 
				(bdd_ithvar((*idMap.find((temp.getLHS()))).second) & bdd_ithvar(i+1)));	
	}
	else {
	  // f = g R h  
	  // x'f <-> (x'h & (x'g | xf)) 
	  transrel &= bdd_biimp(bdd_ithvar(i), bdd_ithvar((*idMap.find((temp.getRHS()))).second) & 
				(bdd_ithvar((*idMap.find((temp.getLHS()))).second) | 
				 bdd_ithvar(i+1)));
	}
	break;
      }
    case Formula::fTemporalUnOp:
      {
	class TemporalUnOp &temp=static_cast<TemporalUnOp &>(*f);
	  if(temp.getOp() == TemporalUnOp::Next) {
	    // f = X g 
	    // x'f <-> xg 
	    transrel &= bdd_biimp(bdd_ithvar(i), bdd_ithvar((*idMap.find((temp.getOperand()))).second+1));		
	  }
	  else if(temp.getOp() == TemporalUnOp::Finally) {
	    transrel &= bdd_biimp(bdd_ithvar(i), 
				  bdd_ithvar((*idMap.find((temp.getOperand()))).second) | bdd_ithvar(i+1));
	  }
	  else if (temp.getOp() == TemporalUnOp::Globally){	
	    transrel &= bdd_biimp(bdd_ithvar(i),  
				  bdd_ithvar((*idMap.find((temp.getOperand()))).second) & bdd_ithvar(i+1));
	  }      
	break;
      }
    case Formula::fBinOp: {
      class BinOp &temp=static_cast<BinOp &>(*f);
      if(temp.getOp() == BinOp::And) {
	// f = g & h 
	// x'f <-> (x'g & x'h)
	//unsigned temp1=(*idMap.find((temp.getLHS()))).second;
	//unsigned temp2=(*idMap.find((temp.getRHS()))).second;
	transrel &= bdd_biimp(bdd_ithvar(i), bdd_ithvar((*idMap.find((temp.getLHS()))).second) & 
			      bdd_ithvar((*idMap.find((temp.getRHS()))).second));
	
      }
      else if(temp.getOp() == BinOp::Or) {
	// f = g | h 
	// x'f <-> (x'g | x'h) 
	transrel &= bdd_biimp(bdd_ithvar(i),  bdd_ithvar((*idMap.find((temp.getLHS()))).second) | 
			      bdd_ithvar((*idMap.find((temp.getRHS()))).second));
      }
      else {	
	return bdd_false();
      }
      break;      
    }
    case Formula::fAtom: {
      break;
    }
    case Formula::fNot: 
      {
	//f = \neg p, p\in AP
	//f <-> \neg p(i)
	class Not &temp=static_cast<Not &>(*f);
	transrel &= bdd_biimp(bdd_ithvar(i), bdd_nithvar((*idMap.find((temp.getOperand()))).second));
      }
      break;
    case Formula::fConst:      
      if(static_cast<Const &>(*f).getVal())
	//f= true
	//f <-> bdd(i) 
	transrel &= bdd_ithvar(i);
      else
	//f=false
	//f <-> !bdd(i)
	transrel &= bdd_nithvar(i);	
      break;      
      //default:
      //assert(false);
    }
  }
  return transrel;
}


/**Check which variables can be elminated by quantification,
 * designate bdd variable numbers to the subformulas, and determine
 * the symbolic intial state
 * @param f Formula being considered
 * @param idMap Table for mapping subformulas to variables
 * @param unquantVar (output) bdd with the remaining variables
 * @param inital (output) the symbolic initial state
 * @param N the number of subformulas
 *@return bdd with the eliminateble variables
 */

static bdd
varEliminate(const class Formula &formula, IdMap &idMap, bdd &unquantVar, bdd &initial, unsigned N) 
{
  //first identify which are to be quantified
  // 2*N+1 = don't quantify, 2*N = quantify
  //flag indicating if previous formula was X (Next operator)
  bool prevN = false;
  for(Formula::Iterator iter=formula.newIterator(); !iter.atEnd(); ++iter) {    
    Formula::Type kind = (*iter).getType();
    switch(kind) { 
    case Formula::fTemporalUnOp:
      if(static_cast<TemporalUnOp &>(*iter).getOp()==TemporalUnOp::Next) {
	prevN=true;
	idMap[&(*iter)]=2*N+1;
      }
      else { 
	prevN=false;
	idMap[&(*iter)]=2*N;
      }
      break;
    case Formula::fTemporalBinOp:
      prevN=false;
      idMap[&(*iter)]=2*N+1;
      break;
    case Formula::fConst:
    case Formula::fNot:
    case Formula::fAtom:
    case Formula::fBinOp:
      if (prevN)
	idMap[&(*iter)]=2*N+1;
      else if(idMap[&(*iter)] < 2*N)
	idMap[&(*iter)]=2*N;
      prevN=false;
      break;
    }
  }

  class bdd vars = bdd_true();
  int varid=0;  
  for(Formula::Iterator iter=formula.newIterator(); !iter.atEnd(); ++iter) {
    if (idMap[&(*iter)] >= 2*N) {
      Formula::Type kind = (*iter).getType();
      switch(kind) { 
      case Formula::fTemporalBinOp:
      case Formula::fTemporalUnOp:
	unquantVar &= bdd_ithvar(varid);
	initial &= bdd_nithvar(varid+1);
	idMap[&(*iter)]=varid;	
	break;
      case Formula::fConst:
      case Formula::fNot:
      case Formula::fAtom:
      case Formula::fBinOp:
	if(varid==0 || idMap[&(*iter)]==2*N+1 ) { //removed 
	  unquantVar &= bdd_ithvar(varid);
	  initial &= bdd_nithvar(varid+1);
	  idMap[&(*iter)]=varid;
	  break;
	}
	idMap[&(*iter)]=varid;
	vars &= bdd_ithvar(varid); // & bdd_ithvar(varid+1);
	break;
      }
      varid+=2;
    }
  }
  return vars;
}



 /**Create an NFA corresponding to a LTL formula in positive normal form
   * @param formula The LTL formula in positive normal form
   * @return Automaton if successful, otherwise null
   */
Automata *
Automata::create(const class Formula &formula) 
{
  const unsigned N=count(formula);
  //init the bdd package
  if (!bdd_isrunning()) bdd_init(500000, 50000);
  if(static_cast<unsigned>(bdd_varnum()) < 2*N) bdd_setvarnum(N*2);
  
  //init variables 
  int *nvar = new int[N];
  int *pvar = new int[N];
  for (unsigned n=0 ; n<N; n++) {
    nvar[n] = n*2;  //normal vars are even
    pvar[n] = n*2+1; //primed vars are odd
  }
  class bdd normvar = bdd_makeset(nvar, N);
  class bdd primvar = bdd_makeset(pvar, N);
  bddPair *backwardPair = bdd_newpair();
  bdd_setpairs(backwardPair, nvar, pvar, N); //we're going backwards
  
  //map formulas to bdd variable numbers
  IdMap idMap;
  class bdd initial=bdd_true();
  class bdd unquantVar=bdd_true();
  class bdd quantVar=varEliminate(formula, idMap, unquantVar, initial, N);
  
  //create transrel
  class bdd transrel = transitions(idMap);
  class bdd transrel2=bdd_exist(transrel, quantVar);

  //compute backward reachability  
  class bdd R = reachable_states(initial, transrel2, primvar, backwardPair);
  bddPair *forwardPair = bdd_newpair();
  bdd_setpairs(forwardPair, pvar, nvar, N); 
  R=bdd_replace(R, forwardPair); //set R to unprimed vars
  
  unsigned sated=static_cast<unsigned>(bdd_satcount(R));
  class bdd test= R & bdd_ithvar(0);
  //forward reachability
  initial = bdd_true() & bdd_ithvar(0);
  class bdd Rprime=reachable_states(initial, transrel2, normvar, forwardPair);  
  R &= Rprime;  

  //clean up
  delete[] nvar;
  delete[] pvar;
  bdd_freepair(backwardPair);
  bdd_freepair(forwardPair);
  if(R == bdd_false()) return new class Automata(formula, bdd_false(), normvar, quantVar, unquantVar, 1u);; 
  return new class Automata(formula, transrel, R, quantVar, unquantVar, 
			    static_cast<unsigned>(bdd_satcount(R&quantVar&primvar))); 
}


#include "NonDetAut.h"

NonDetAut *
Automata::toNonDetAut(IntBddMap &labels, AtomMap &idmap) const {
  const unsigned N=count(myFormula);
  int *nvar = new int[N];
  int *pvar = new int[N];
  class bdd final=bdd_true();
  for(unsigned i=0; i<N; i++) {
    nvar[i]=2*i;
    pvar[i]=2*i+1;
    final &= bdd_nithvar(2*i);
  }
  class bdd vars = bdd_makeset(nvar, N);
  class bdd prvar = bdd_makeset(pvar,N); 
  bddPair *backwardPair = bdd_newpair();
  bddPair *forwardPair = bdd_newpair();
  bdd_setpairs(backwardPair, nvar, pvar, N);
  bdd_setpairs(forwardPair, pvar, nvar, N);

  //map bdd variable numbers to ap IDs
  class bdd notatomic=bdd_true();
  idmap.reserve(2*N);
  {
    unsigned i=0;
    FormulaSet fTable;
    for (class Formula::Iterator t=myFormula.newIterator(); !t.atEnd(); ++t) {
      Formula::Type kind=(*t).getType();
      if(fTable.find(&(*t)) == fTable.end()) {
	fTable.insert(&(*t));
	if (kind == Formula::fAtom) { 
	  idmap[2*i]=static_cast<const class Atom &>(*t);
	}
	else {
	  notatomic &= bdd_ithvar(2*i);
	}	
	i++;
      }
    }
    assert(i==N);
  }

  NonDetAut *result=new NonDetAut(mySize, 0);
  NonDetAut &nondetaut=*result; 
  class bdd apTransRel=bdd_exist(myTransRel, myQuantified);
  class bdd apSourceSpace = myStateSpace;
  unsigned state=0;
  unsigned arcCount=0;
  BddMap bddMap;
  BddMap bddArcMap;
  for(class bdd source=bdd_satoneset(apSourceSpace, myUnQuantified, bdd_false());
      apSourceSpace  != bdd_false(); 
      apSourceSpace &= !source, 
	source=bdd_satoneset(apSourceSpace, myUnQuantified, bdd_false())) {
    BddMap::const_iterator iter=bddMap.find(source);
    assert((source & myStateSpace) != bdd_false());    
    if(iter==bddMap.end()) { //print state number
      bddMap[source]=BddMap::data_type(state++, false);
      //bdd_printset(source); fputs("\n", stdout);
    }
    if ( (source & bdd_ithvar(0)) != bdd_false()) // is it an initial state
      nondetaut.makeInitial(bddMap[source].first);
    if ((source & final) != bdd_false()) //is it the final state 
      nondetaut.makeFinal(bddMap[source].first);
    //create arcs 
    class bdd targetSpace=bdd_relprod(source, apTransRel, vars);
    targetSpace=bdd_replace(targetSpace, forwardPair);
    for(class bdd target=bdd_satoneset(targetSpace, myUnQuantified, bdd_false());
	targetSpace != bdd_false(); targetSpace &= !target,  
      target=bdd_satoneset(targetSpace, myUnQuantified, bdd_false())) { 
      if ((target & myStateSpace) == bdd_false()) continue;
      class bdd primtarget=bdd_replace(target, backwardPair);
      BddMap::const_iterator targetIter=bddMap.find(target);
      if(targetIter==bddMap.end()) {
	bddMap[target]=BddMap::data_type(state++, false);
	//bdd_printset(target); fputs("\n", stdout);
      } 
      //compute the arc between the states       
      class bdd arc=source&myTransRel&primtarget;
      assert(arc != bdd_false());
      arc=bdd_exist(arc, prvar);
      arc=bdd_exist(arc, notatomic);
      //bdd_fprintset(stderr, arc);      
      if(bddArcMap.find(arc) == bddArcMap.end()) {
	labels.push_back(arc);
	bddArcMap[arc]=BddMap::data_type(arcCount++, false);
	if(nondetaut.alphabetSize() <  arcCount )
	  nondetaut.setASize(arcCount);
      }
      nondetaut.addTransition(bddMap[source].first, bddArcMap[arc].first, bddMap[target].first);
    }
  }

  delete[] nvar;
  delete[] pvar;
  bdd_freepair(backwardPair);
  bdd_freepair(forwardPair);
  return result;
}







