#include <cassert>
#include "Buchi.h"
#include "Formula.h"
#include "TemporalBinOp.h"
#include "TemporalUnOp.h"
#include "BinOp.h"
#include "Not.h"
#include "Atom.h"
#include "Const.h"
#include "NonDetAut.h"

#include "BddPrint.h" 



/**Compute el(f)
 *@param f The input formula
 *@param set (output) Place holder for el(f)
 */
static void
computeEL(const class Formula *f, FormulaMap &set)
{
  for(Formula::Iterator iter=f->newIterator(); !iter.atEnd(); ++iter) {
    Formula::Type kind=(*iter).getType();
    switch (kind) {
    case Formula::fTemporalBinOp: {
      const class TemporalBinOp &temp=static_cast<const TemporalBinOp &>(*iter);
      assert(temp.getOp()==TemporalBinOp::Until || temp.getOp()==TemporalBinOp::Release);      
      class TemporalUnOp *nextedFormula=new TemporalUnOp(TemporalUnOp::Next, &temp);
      if(set.find(nextedFormula) == set.end())
	set.insert(FormulaMap::value_type(nextedFormula, set.size()));
      else
	delete nextedFormula;
      //f(set.find(f) == set.end())
      //  set.insert(FormulaMap::value_type(f, set.size()));
      break;
    }
    case Formula::fTemporalUnOp: {
      const class TemporalUnOp  &temp=static_cast<const TemporalUnOp &>(*iter);
      if(temp.getOp()==TemporalUnOp::Next) {
	if(set.find(&temp) == set.end())
	  set.insert(FormulaMap::value_type(&temp, set.size()));
	break;
      }
      class TemporalUnOp *nextedFormula=new TemporalUnOp(TemporalUnOp::Next, &temp);
      if(set.find(nextedFormula) == set.end())
	set.insert(FormulaMap::value_type(nextedFormula, set.size()));
      else
	delete nextedFormula;    
      break;
    }
    case Formula::fBinOp: {
      const class BinOp &temp=static_cast<const BinOp &>(*iter);
      assert(temp.getOp()==BinOp::Or ||  temp.getOp()==BinOp::And);
      break;
    }
    case Formula::fAtom: {
      if(set.find(&(*iter)) == set.end())
	set.insert(FormulaMap::value_type(&(*iter), set.size()));;
      break;
    }
    case Formula::fNot: {
      //const class Not &temp=static_cast<const Not &>(*iter);
      break;
    }
    case Formula::fConst: {
      break;
    }
    default:    
      assert(false);
    }
  }
}


/**
 *@param f The input formula
 *@param set The set el(f)
 *@return bdd representing the set of states satisfying f
 */
static bdd
sat(const class Formula *f, const FormulaMap &set) 
{
  if(set.find(f) != set.end()) { //take care of elementary formulas
    //if(f->getType()==Formula::fAtom && (!static_cast<const Atom *>(f)->getSign())) { //buggy!
     //  return bdd_nithvar(2*((*set.find(f)).second));
    //}
    return bdd_ithvar(2*((*set.find(f)).second)); 
  }

  FormulaBddMap bddmap;
  class bdd result=bdd_true();
  for(Formula::PostIterator iter=f->newPostIterator(); !iter.atEnd(); ++iter) {
    FormulaMap::const_iterator i=set.find(&(*iter));
    //only even (unprimed) var nums are returned
    Formula::Type kind=(*iter).getType();
    if(i != set.end()) { //take care of elementary formulas
      //      if(kind==Formula::fAtom && (!static_cast<const Atom &>(*iter).getSign())) { //buggy!
      //	result=bdd_nithvar(2*((*i).second));
      //} 
      //else {
	result=bdd_ithvar(2*((*i).second)); 
	//}
      bddmap[&(*iter)]=result;
      continue;
    }
    
    switch(kind) {      
    case Formula::fNot: { 
      const class Not &temp = static_cast<const Not &>(*iter);
      assert(bddmap.find(temp.getOperand())!=bddmap.end());
      result=!bddmap[temp.getOperand()]; //Possible bug, don't cares?!? 
      bddmap[&temp]=result;
      break;
    } 
    //case f= g & h or f = g | h
    case Formula::fBinOp: { 
      const class BinOp & temp = static_cast<const BinOp &>(*iter);
      if(temp.getOp()==BinOp::Or) { 
	result = bddmap[temp.getLHS()] |  bddmap[temp.getRHS()];
      }
      else if (temp.getOp() == BinOp::And) {
	result = bddmap[temp.getLHS()] &  bddmap[temp.getRHS()];
      }
      else {
	assert(false);
	result = bdd_false(); 
      }
      bddmap[&temp]=result;
      break;
    } 
    //case f = g U h
    case Formula::fTemporalBinOp: {
      const class TemporalBinOp &temp = static_cast<const class TemporalBinOp &>(*iter);
      class TemporalUnOp *nexted = new TemporalUnOp(TemporalUnOp::Next, &temp);
      FormulaMap::const_iterator nextiter=set.find(nexted);
      assert(nextiter!=set.end());
      if(temp.getOp()==TemporalBinOp::Until) { 
	result =  bddmap[temp.getRHS()] | ( bddmap[temp.getLHS()] & bdd_ithvar(2*((*nextiter).second)));
      }
      else if (temp.getOp() == TemporalBinOp::Release) {
	result =  bddmap[temp.getRHS()] & ( bddmap[temp.getLHS()] | bdd_ithvar(2*((*nextiter).second)));
      }
      else {
	assert(false);
	result= bdd_false(); 
      }
      delete nexted;
      bddmap[&temp]=result;
      break;
    }
    case Formula::fTemporalUnOp: {
      const class TemporalUnOp &temp = static_cast<const class TemporalUnOp &>(*iter);
      class TemporalUnOp *nexted = new TemporalUnOp(TemporalUnOp::Next, &temp);
      FormulaMap::const_iterator nextiter=set.find(nexted);
      assert(nextiter!=set.end());
      if(temp.getOp()==TemporalUnOp::Finally) { 
	result =  bddmap[temp.getOperand()] | bdd_ithvar(2*((*nextiter).second));
      }
      else if (temp.getOp() == TemporalUnOp::Globally) {
	result =  bddmap[temp.getOperand()] & bdd_ithvar(2*((*nextiter).second));
      }
      else {
	assert(false);
	result= bdd_false(); 
      }
      delete nexted;
      bddmap[&temp]=result;
      break;
    }
    case Formula::fConst: {    
      const class Const &temp =static_cast<const class Const &>(*iter);
      if(temp.getVal())
	result= bdd_true();
      else 
	result=bdd_false();
      bddmap[&temp]=result;
      break;
    }  
    default:
      assert(false);    
    }
  }
  return result;
}

/**Make the transition relation
 *@param set The set of elmentary formulas
 *@param replacePair bdd pair for replacing unprimed with primed vars
 *@return the transition relation for the bdd
 */
static bdd
makeTransRel(const FormulaMap &set, bddPair *replacePair)
{
  bdd transrel = bdd_true();
  
  for(FormulaMap::const_iterator f=set.begin(); f!=set.end(); ++f) {
    Formula::Type kind=(*f).first->getType();
    if(kind != Formula::fTemporalUnOp) continue;
    const class TemporalUnOp *temp=static_cast<const TemporalUnOp *>((*f).first);
    assert(temp->getOp() == TemporalUnOp::Next);
    class bdd primed=bdd_replace(sat(temp->getOperand(), set), replacePair);
    transrel &= bdd_biimp(sat(temp, set), primed);
  }  
  return transrel;
}

/**Compute the acceptance sets
 *@param f The formula of the automaton
 *@param bddList (output) Storage for the acceptance sets
 *@param elSet Table mapping the elementary formulas to bdd variables 
 */

static void
fairSet (const class Formula &f, bddList &fairset, FormulaMap &elSet)
{
  for(Formula::Iterator i=f.newIterator(); !i.atEnd(); ++i) {
    if((*i).getType() == Formula::fTemporalBinOp) {
      const class TemporalBinOp &temp=static_cast<const TemporalBinOp &>(*i);
      if(temp.getOp() == TemporalBinOp::Until) {
	class BinOp *fair=new BinOp(BinOp::Or, new Not(&temp), temp.getRHS());      
	bdd result=sat(fair, elSet);
	fairset.push_front(result);
	delete fair->getLHS();
	delete fair;
      }
    }
    if ((*i).getType() == Formula::fTemporalUnOp) {
      const class TemporalUnOp &temp=static_cast<const TemporalUnOp &>(*i);
      if(temp.getOp() == TemporalUnOp::Finally) {
	class BinOp *fair=new BinOp(BinOp::Or, new Not(&temp), temp.getOperand());      
	bdd result=sat(fair, elSet);
	fairset.push_front(result);
	delete fair->getLHS();
	delete fair;
      }
    }
  }
}

inline static unsigned 
count(const class Formula &f) {
  FormulaMap fSet;
  unsigned count=0;
  for(class Formula::Iterator t=f.newIterator(); !t.atEnd(); ++t) {
    if (fSet.find(&(*t)) == fSet.end()) { 
      fSet.insert(FormulaMap::value_type(&(*t), count));
      count++;
    }
  }
  return count;
}

/**
 *@param The formula under inspection
 *@return A bdd with the atomic propositions conjuncted
 */
bdd apBDD(const class Formula *f) 
{
  FormulaMap map;
  computeEL(f, map);

  class bdd ap=bdd_true();
  
  for (FormulaMap::const_iterator t=map.begin(); t!=map.end(); ++t) {
    Formula::Type kind=(*t).first->getType();
    if (kind == Formula::fAtom) { 
      ap &= bdd_ithvar(2*((*t).second));
    }	
  }
  return ap;
}

class Buchi*
Buchi::create(const class Formula &formula)
{
  FormulaMap map;
  computeEL(&formula, map);
  unsigned N=map.size();
  if (!bdd_isrunning()) bdd_init(500000, 50000);
  if (static_cast<unsigned>(bdd_varnum()) < 2*N) bdd_setvarnum(2*N);
  
  int *nvar = new int[N];
  int *pvar = new int[N];
  for (unsigned n=0 ; n<N; n++) {
    nvar[n] = n*2;  //normal vars are even
    pvar[n] = n*2+1; //primed vars are odd
  }

  class bdd normvar = bdd_makeset(nvar, N);
  class bdd primvar = bdd_makeset(pvar, N);
  bddPair *replacePair = bdd_newpair();
  bdd_setpairs(replacePair, nvar, pvar, N); 
  bdd transrel=makeTransRel(map, replacePair);
  bddList *fairness= new bddList();
  fairSet(formula, *fairness, map);

  bddPair *forwardPair = bdd_newpair();
  bdd_setpairs(forwardPair, pvar, nvar, N); 
  class bdd initial=sat(&formula, map);
  class bdd statespace = initial; // Reachable state space
  class bdd prevR; // Previously reached state space and temp
  do  {
     prevR = statespace;
     class bdd newset  = bdd_appex(statespace, transrel, bddop_and, normvar);
     newset = bdd_replace(newset, forwardPair);
     statespace |= newset;
  }
  while(prevR != statespace);
 
  delete[] nvar;
  delete[] pvar;
  bdd_freepair(replacePair);
  bdd_freepair(forwardPair);
  return new class Buchi(formula, transrel, statespace, fairness, initial);
}

Buchi::Buchi(const class Formula &f, class bdd &transrel, class bdd &statespace, 
	     bddList *fairness, class bdd &initial) : 
	     myFormula(f), myTransRel(transrel), myStateSpace(statespace), myFairness(fairness), 
	     myInitial(initial)
	     
{
  ;
}

Buchi::~Buchi() 
{
  delete myFairness;
}



#include "ExplicitBuchi.h"

class ExplicitBuchi *
Buchi::toExplicit(IntBddMap &labels, AtomMap &idmap) const
{
  FormulaMap map;
  computeEL(&myFormula, map);
  unsigned N=map.size();

  int *nvar = new int[N];
  int *pvar = new int[N];
  for(unsigned i=0; i<N; i++) {
    nvar[i]=2*i;
    pvar[i]=2*i+1;
  }
  class bdd vars = bdd_makeset(nvar, N);
  class bdd prvar = bdd_makeset(pvar,N); 
  bddPair *backwardPair = bdd_newpair();
  bddPair *forwardPair =bdd_newpair();
  class bdd notatomic=bdd_true();
  bdd_setpairs(backwardPair, nvar, pvar, N);
  bdd_setpairs(forwardPair, pvar, nvar, N);

  //find not atomic elementary formulas
  idmap.reserve(2*N);
  for (FormulaMap::const_iterator t=map.begin(); t!=map.end(); ++t) {
    Formula::Type kind=(*t).first->getType();
    if (kind != Formula::fAtom) { 
      notatomic &= bdd_ithvar(2*((*t).second));
    }	
    else {
      idmap[2*((*t).second)]=dynamic_cast<const class Atom &>(*(*t).first);
    }
  }

  //corect size expression
  unsigned size = static_cast<unsigned>(bdd_satcount(myStateSpace&prvar));  
  ExplicitBuchi *result=new ExplicitBuchi((size> (1U<<N))? (1U<<N) + 1 : size + 1, 
					  0, myFairness->size());
  class bdd apSourceSpace = myStateSpace;
  unsigned state=1;
  unsigned arcCount=0;
  BddMap bddMap;
  BddMap bddArcMap; //map bddlabels to label id:s
  for(class bdd source=bdd_satoneset(apSourceSpace, vars, bdd_false());
      apSourceSpace != bdd_false(); 
      apSourceSpace &= !source, 
	source=bdd_satoneset(apSourceSpace, vars, bdd_false())) {
    BddMap::const_iterator iter=bddMap.find(source);
    if(iter==bddMap.end()) { //record state number
      bddMap[source]=BddMap::data_type(state++, false);
    }
    //bool initial=false;
    if ( (source & myInitial) != bdd_false()) // is it an initial state
      //initial=true; //      
      result->makeInitial(bddMap[source].first);
    unsigned fair=myFairness->size()-1;
    for (bddList::const_iterator iter=myFairness->begin(); iter!=myFairness->end(); ++iter, --fair) {
      if ((source & *iter) != bdd_false()) 
	result->makeFinal(bddMap[source].first, fair); 
    }
   //create arcs 
    class bdd targetSpace=bdd_relprod(source, myTransRel, vars);
    targetSpace=bdd_replace(targetSpace, forwardPair);
    for(class bdd target=bdd_satoneset(targetSpace, vars, bdd_false());
	targetSpace != bdd_false(); targetSpace &= !target, 
	  target=bdd_satoneset(targetSpace, vars, bdd_false())) {      
      if( (target & myStateSpace) == bdd_false()) continue;
      BddMap::const_iterator targetIter=bddMap.find(target);
      if(targetIter==bddMap.end()) {
	bddMap[target]=BddMap::data_type(state++, false);
      } 
      //compute the arc between the states       
      class bdd arc=bdd_exist(source, notatomic);
      if(bddArcMap.find(arc) == bddArcMap.end()) {
	labels.push_back(arc);
	bddArcMap[arc]=BddMap::data_type(arcCount++, false);	
	if(result->alphabetSize() <  arcCount )
	  result->setASize(arcCount);
      }
      result->addTransition(bddMap[source].first, bddArcMap[arc].first, bddMap[target].first);
      //if(initial) result->addTransition(0, bddArcMap[arc].first, bddMap[target].first);
    }
  }
  delete[] nvar;
  delete[] pvar;
  bdd_freepair(backwardPair);

  //transform the NFA to an automaton with only one initial state
  for(unsigned i=result->size(); i--; ) {
    if(result->isInitial(i)) {
      for (unsigned j=result->alphabetSize(); j--; ) { 
	for(unsigned k=result->numArcs(i,j); k--; ) {
	  result->addTransition(0, j, result->dest(i,j,k));
	}
      }
    }
    result->delInitial(i);
  }
  result->makeInitial(0);  
  return result;
}

//    for(unsigned target=result->size(); target--; ) {
//      class bdd arc=bdd_false();
//    for(unsigned p=result->size(); p--; ) {
//      if(result->isInitial(p)) {
//        for (unsigned q=result->alphabetSize(); q--; ) { 
//  	for(unsigned r=result->numArcs(p,q); r--; ) {
//  	  if(target==result->dest(p,q,r)) {		    
//  	    arc |= labels[q];		    
//  	  }
//  	}	  
//        }
//      }
//    }
//    if ( arc!= bdd_false()) {
//      if(bddArcMap.find(arc) == bddArcMap.end()) {
//        labels.push_back(arc);
//        bddArcMap[arc]=BddMap::data_type(arcCount++, false);	
//        if(result->alphabetSize() <  arcCount )
//  	result->setASize(arcCount);
//      }
//      result->addTransition(0, bddArcMap[arc].first, target);
//    }
//    }
//    for(unsigned i=result->size(); --i; ) result->delInitial(i);
//    result->makeInitial(0);
//    return result;
//}

