#pragma implementation
#include "DetAut.h"
#include "NonDetAut.h"
#include <cstdlib>

DetAut::DetAut(unsigned size, unsigned aSize)
 : AbsAutomata(size, aSize), myAllocated(size) {
  assert(size > 0);
  myTransRel=new unsigned[size*aSize];
  memset(myTransRel, 0, size*aSize * sizeof(unsigned));
}

void
DetAut::setSize(unsigned size) 
{
  myInitial.setSize(size);
  myAccepting.setSize(size);
  mySize=size; 
  if(size <= myAllocated) {
    return;
  }
  unsigned oldAllocated=myAllocated;
  while(size>=myAllocated) myAllocated *= 2; 
  unsigned *temp=new unsigned[myAllocated*myAlphabetSize];
  memcpy(temp, myTransRel, oldAllocated*myAlphabetSize * sizeof(unsigned));
  memset(temp+oldAllocated*myAlphabetSize, 0, 
	 (myAllocated-oldAllocated)*myAlphabetSize * sizeof(unsigned));
  delete[] myTransRel;
  myTransRel=temp;
}


#include <pair.h>
#include <hash_set>
#include "BitVector.h"

struct eqset
{
  bool operator()(const class BitVector &s1, const class BitVector &s2) const
  {
    return s1==s2;
  }
};
struct hasher
{
  unsigned operator()(const class BitVector &s) const
  {
    unsigned hvalue=246093032;
    return s.hash(hvalue);
  }
};
typedef std::hash_set<class BitVector, hasher, eqset> PowerSet;

typedef std::pair<class BitVector, unsigned> Pair;
struct eqpair {
  bool operator()(const Pair &p1, const Pair &p2) const {
    return p1.first==p2.first && p1.second==p2.second;
  }
};
struct phasher 
{
  unsigned operator()(const Pair &s) const
  {
    unsigned hvalue=246093032;
    hvalue ^= s.first.hash(hvalue);
    return hvalue^(s.second);
  }
};
typedef std::hash_set<Pair, phasher, eqpair> PairSet;

/**@return Can the two sets be split w.r.t. to a certain label*/
inline static bool
splittable(const class DetAut &aut, const class BitVector &q0, 
	   const class BitVector &q1, unsigned label) 
{
  for(unsigned p=q0.getSize(); p--; )  {
    if(q0[p]) {
      for(unsigned q=p; q--; ) {
	if(q0[q]) {	  
	  if( ((aut.numArcs(p, label) && q1[aut.dest(p, label)]) && 
	       (!aut.numArcs(q, label) || !q1[aut.dest(q, label)])) ||
	      ((!aut.numArcs(p, label) || !q1[aut.dest(p, label)]) && 
	       (aut.numArcs(q, label) && q1[aut.dest(q, label)])) )
	    return true;
	}
      }
    }
  }
  return false;
} 

/**Compute a relative transitive closure
 *@param aut Automaton concerned
 *@param q0 Set of start states
 *@param q1 All dests must be in this set
 *@result place holder for the result
 */
inline static void
closure(const DetAut &aut, const class BitVector &q0, 
	const class BitVector &q1, unsigned label, class BitVector &result)
{
  for (unsigned p=q0.getSize(); p--; ) {
    if(q0[p] && aut.numArcs(p, label) && q1[aut.dest(p,label)])
      result.assign(p, true);
  }
}
/**Remove a sink state from a complete deterministic automaton
 *@param result Automaton to be processed
 *@param True Index for the "true"-label in the alphabet.   
 * If True>=result.alphabetSize() the automaton has no "true"-label.
 */
void
removeSink(class DetAut &result, unsigned True) 
{
  //check that automaton has "true"-label
  if(True>=result.alphabetSize()) return; 

//check for a sink state 
  unsigned sink=0;
  bool flag=false;
  for(unsigned i=result.size(); i--  && !flag; ) {
    if(result.numArcs(i, True)==0 || result.dest(i, True)!=i
       || result.isInitial(i) || result.isFinal(i))
      continue;    
    for(unsigned j=result.alphabetSize(); j-- ; ) {
      if(j!=True && result.numArcs(i,j)!= 0)
	break;
      if(j==0) {//i is a sink
	sink=i;
	flag=true;
      }
    }
  }
  //remove the sink state
  if(flag) {
    for(unsigned i=0; i<sink; i++) {
 	for(unsigned j=result.alphabetSize(); j--; ) {
	  if (result.numArcs(i,j)==0) continue;
	  if(result.dest(i,j) > sink)	//implicit delete     	    
	    result.addTransition(i, j, result.dest(i,j) - 1);
	  else if(result.dest(i,j) == sink)
	    result.delTransition(i, j); 
	}
    }    
    //move the transitions of the states with an index 
    //greater than the sink to one smaller index
    for(unsigned i=sink; i<result.size()-1; i++) {
      if(result.isFinal(i+1)) {
	result.delFinal(i+1);
	result.makeFinal(i);
      }
      if(result.isInitial(i+1)) {
	result.delInitial(i+1);
	result.makeInitial(i);
      }
      for(unsigned j=result.alphabetSize(); j--; ) {
	//delete the old transition
	result.delTransition(i, j);
	if (result.numArcs(i+1,j)==0) continue;
	if(result.dest(i+1, j)<sink)
	    result.addTransition(i, j, result.dest(i+1, j));
	else if(result.dest(i+1,j)==sink)
	  result.delTransition(i, j);
	else
	  result.addTransition(i, j, result.dest(i+1, j) - 1);
      }
    }
    result.setSize(result.size() - 1);
  }
}

void 
DetAut::setAlphabetSize(unsigned size)
{
  if(size<=myAlphabetSize) return;
  unsigned *temp=new unsigned[myAllocated*size];
  for(unsigned i=myAllocated; i--; ) {
    memcpy(&temp[i*size], &myTransRel[i*myAlphabetSize], myAlphabetSize * sizeof(unsigned));
    temp[i*size + size-1]=0;
  }   
  delete[] myTransRel;
  myTransRel=temp;
  myAlphabetSize=size;
}



/**Minimize the deterministic automaton 
 *@return A minimized DetAut
 */
DetAut*
DetAut::minimize() const
{
  PowerSet P;
  //init P
  P.insert(myAccepting);  
  BitVector allStates(mySize); 
  for(unsigned i=allStates.getSize(); i--; )
    allStates.assign(i, true);
  allStates.andNot(myAccepting);
  P.insert(allStates);

  //init L
  PairSet L; 
  if(myNumFinal <= (mySize - myNumFinal)) 
    for (unsigned i=myAlphabetSize; i--; )
      L.insert(PairSet::value_type(myAccepting,i)); 
  else {
    for (unsigned i=myAlphabetSize; i--; )
      L.insert(PairSet::value_type(allStates,i));
  }
  
  while(!L.empty()) {
    Pair pair = *L.begin();
    L.erase(L.begin());
    const BitVector &q1=pair.first;
    PowerSet POld(P);
    for(PowerSet::const_iterator iter=POld.begin(); iter!=POld.end(); ++iter) {
      const BitVector &q0=*iter;
      if (!splittable(*this, q0, q1, pair.second))
	continue;
      BitVector q0prime(mySize);
      closure(*this, q0, q1, pair.second, q0prime);
      P.erase(q0);
      BitVector q0Minus(q0);
      q0Minus.andNot(q0prime);
      P.insert(q0prime);
      P.insert(q0Minus);      
      for(unsigned i=0; i<myAlphabetSize; ++i) {
	PairSet::iterator j=L.find(Pair(q0, i));
	if (j != L.end()) {
	  L.erase(j);
	  L.insert(Pair(q0prime, i));
	  L.insert(Pair(q0Minus, i));
	}
	else {
	  if (q0prime.numTrue() <= q0Minus.numTrue())
	    L.insert(Pair(q0prime, i)) ;
	  else
	    L.insert(Pair(q0Minus, i));
	}
      }    
    }
  }
   
  //construct the reduced Automaton
  unsigned size=P.size();
  class DetAut &result=*new DetAut(size, myAlphabetSize);
  unsigned source=0;
  for (PowerSet::const_iterator i=P.begin(); i!=P.end(); ++i, source++) {    
    for(unsigned v=myAlphabetSize; v--; ) {
      unsigned dest=0;  
      for (PowerSet::const_iterator j=P.begin(); j!=P.end(); ++j, dest++) {
	for(unsigned p=(*i).getSize(); p--; ) {
	  if ((*i)[p]) {
	    if(myAccepting[p])
	      result.makeFinal(source);
	    if(myInitial[p])
	      result.makeInitial(source);		
	    if ((*j)[(this->dest(p, v))]) {	       
	      result.addTransition(source, v, dest);
	      break;
	    }
	  }
	}
      }
    }
  }   
  removeSink(result, 0); //hack, does not work!!!
  return &result;
}


//We follow Vardi's version of Buchi complementation
NonDetAut *
DetAut::buchiComplement(unsigned True) const
{
  NonDetAut *result=new NonDetAut(2*mySize, (True==myAlphabetSize) ? True+1 : myAlphabetSize);
  
  for(unsigned state=mySize; state--;) {
    if(this->isInitial(state)) result->makeInitial(state);
    if(this->isFinal(state)) {
      result->addTransition(state, True, state);            
      continue;
    }    
    else {
      result->makeFinal(mySize+state);    	  
    }
    for(unsigned action=myAlphabetSize; action--; ) {
      if(!(this->numArcs(state, action))) continue;
      result->addTransition(state, action, this->dest(state, action));
      if(!this->isFinal(this->dest(state, action))) {
	result->addTransition(state, action, mySize+(this->dest(state, action)));
	result->addTransition(mySize+state, action, mySize+(this->dest(state, action)));
      }
    }    
  }
  return result;
}

#include "bdd.h"

NonDetAut *
DetAut::labelJoin(IntBddMap &labels)
{
  NonDetAut *nondetaut=new NonDetAut(mySize, myAlphabetSize);
  NonDetAut &result=*nondetaut;

  unsigned *collision = new unsigned[this->size()];
  unsigned *tempDest = new unsigned[this->alphabetSize()];
  
  for(unsigned source=this->size(); source--; ) {
    if(this->isInitial(source)) result.makeInitial(source);
    if(this->isFinal(source)) result.makeFinal(source);
    for(unsigned i= this->size(); i--;) collision[i]=this->alphabetSize(); //reset
    for(unsigned label=0; label<this->alphabetSize(); label++) {
      //Create union-find structure
      if(!this->numArcs(source, label)) {
	tempDest[label]=label;
	continue;
      }
      if (collision[this->dest(source, label)]!=this->alphabetSize()) {
	tempDest[label]=collision[this->dest(source, label)];
	collision[this->dest(source, label)]=label;
      }
      else {
	tempDest[label]=label;	     
	collision[this->dest(source, label)]=label;	
      }
    }    
    //going in the reverse direction is critical
    for (unsigned label=this->alphabetSize(); label--; ) {
      if (!this->numArcs(source, label)) continue; //skip arcs which do not exist
      if (tempDest[label]==label) { //copy arcs which are singeltons
	result.addTransition(source, label, this->dest(source, label));
	continue;
      }
      class bdd newArc=labels[label]; //bdd of previously encountered label
      unsigned join=label;
      do { 
	unsigned prev=join;
	join=tempDest[join];
	tempDest[prev]=prev;
	this->delTransition(source, join); //delete equivalent arc
	class bdd arc=labels[join]; //bdd of equivalent label
	newArc |= arc; //joined label	
      } while(join!=tempDest[join]);
      //add the new joined label, at this point there will always be at least one
      IntBddMap::iterator iter = labels.begin();
      unsigned index=0;
      for (; iter!=labels.end() && *iter!=newArc; ++iter, ++index);
      if(iter != labels.end()) {	
	result.addTransition(source, index, this->dest(source, label));
	this->delTransition(source, label);
      }
      else {
	labels.push_back(newArc); 
	result.setASize(result.alphabetSize()+1);
	result.addTransition(source, result.alphabetSize()-1, this->dest(source, label));
	this->delTransition(source, label); //delete last equivalent arc 	
      }
    }
  }
  delete[] tempDest;
  delete[] collision;
  return nondetaut;
}
