#pragma implementation
#include "ExplicitBuchi.h"
#include <cstdlib>
#include "BddPrint.h"

ExplicitBuchi::ExplicitBuchi(unsigned size, unsigned aSize, unsigned setsize)
  : mySize(size), myAlphabetSize(aSize), myNumAccSets(setsize),  
  myFinal(size*setsize), myInitial(size), myAllocated(aSize) 
{
  assert(size>0);
  myTransRelSize=new unsigned[size*aSize];
  myTransRel=new unsigned*[size*aSize];
  memset(myTransRelSize, 0, size*aSize * sizeof (unsigned));
  memset(myTransRel, 0, size*aSize * sizeof (unsigned));
}

ExplicitBuchi::~ExplicitBuchi()
{
  for (unsigned i=mySize*myAllocated; i--; ) delete[] myTransRel[i];
  delete[] myTransRel;
  delete[] myTransRelSize;
}


void
ExplicitBuchi::setASize(unsigned size) 
{
  if (size<=myAllocated) {
    myAlphabetSize=size;
    return;
  }
  unsigned oldSize=myAllocated;
  while(size >= myAllocated) {++myAllocated;  myAllocated *= 2;}
  unsigned *temp=new unsigned[mySize*myAllocated];
  memcpy(temp, myTransRelSize, mySize*oldSize*sizeof (unsigned));
  memset(temp + mySize*oldSize, 0, mySize*(myAllocated-oldSize)*sizeof(unsigned)); 
  delete[] myTransRelSize;
  myTransRelSize=temp;
  unsigned **temporary=new unsigned*[mySize*myAllocated];
  memcpy(temporary, myTransRel, mySize*oldSize*sizeof (unsigned));
  memset(temporary + mySize*oldSize, 0, mySize*(myAllocated-oldSize)*sizeof(unsigned)); 
  delete[] myTransRel;
  myTransRel=temporary;
  myAlphabetSize=size;
}

void 
ExplicitBuchi::addTransition(unsigned source, unsigned label, unsigned dest) {
  assert(source<mySize && label<myAlphabetSize && dest<mySize);
  unsigned *temp=new unsigned[myTransRelSize[label*mySize + source]+1];
  memcpy(temp, myTransRel[label*mySize + source],
	 myTransRelSize[label*mySize + source] * sizeof(unsigned));
  delete[] myTransRel[label*mySize + source];
  temp[myTransRelSize[label*mySize + source]++]=dest;
  myTransRel[label*mySize + source]=temp;
}


void
ExplicitBuchi::printAut(FILE *output, const IntBddMap &labels, const AtomMap &idmap) const {
  fprintf(output, "%u\n", this->size());
  fprintf(output, "%u\n", this->numSets());
  for(unsigned i=this->size(); i--; ) {
    fprintf(output, "%u ", i);
    if(this->isInitial(i))
      fputs("1 ", output);
    else 
      fputs("0 ", output);
    if(this->isFinal(i)) {
      for(unsigned k=this->numSets(); k--; ) 
	if (this->final(i, k))
	    fprintf(output, "%u ", k);
    }
   fputs("-1\n", output);
    for(unsigned j=this->alphabetSize(); j--; ) {
      for(unsigned k=this->numArcs(i, j); k--;  ) {
        fprintf(output, "%u ", this->dest(i,j, k));
        printArc(labels[j], idmap, output);
      }
    }
    fputs("-1\n", output);
  }
} 

#include <ctype.h>
#include "BinOp.h"
#include "Not.h"
#include "Const.h"


/**Parse a propositional formula from an input stream. The formula
 * is expected to be in post-fix notation.
 * @param input the input stream
 * @return formula parsed from the stream
 */

static Formula *parseGate(FILE *input) 
{
  int ch;
  //eat whitespace
  while(isspace(ch=fgetc(input)));
  switch(ch) {
  case 'i':
  case 'e':
  case '&':
  case '|':
    if(class Formula *left=parseGate(input)) {
      if(class Formula *right=parseGate(input)) {
        switch(ch) {
        case 'i':
          return new class BinOp(BinOp::Impl, left, right);
        case 'e':
          return new class BinOp(BinOp::Equiv, left, right);
        case '&':
          return new class BinOp(BinOp::And, left, right);
        case '|':
          return new class BinOp(BinOp::Or, left, right);
        }
      }
      left->destroy();
    }
    return 0;
  case '!':
    if(class Formula *operand=parseGate(input)) {
      if(ch=='!') {
        return new class Not(operand);
      }
    }
    return 0;
  case 't':    
  case 'f':
    return new class Const(ch=='t');
  case 'p':    
    {     
      unsigned num;
      if (1 != fscanf (input, "%u", &num))
        fputs ("Error in proposition number.\n", stderr);
      else
        return new class Atom(num);
    }
  case EOF:
    fprintf(stderr, "Parse error. Unexpected end of file.\n");
    return 0;
  default:
    fprintf (stderr, "Parse error. Illegal character %c\n ", ch);
    return 0;
  }
}


/**Create a bdd representation of a propositional formula. The formula and
 * a mapping from atomic propositions to bdd variable numbers is passed to the algorithm 
 * as parameters.
 *@param f the formula
 *@param mapping Mapping from atomic propositions to bdd variable numbers
 *@return bdd representation of the formula
 */

static bdd createLabel(const class Formula *f, const FormulaMap &mapping) 
{ 
  if(mapping.find(f)!=mapping.end()) {    
    return bdd_ithvar((*mapping.find(f)).second);
  }

  FormulaBddMap bddmap;
  bdd result;
  for(class Formula::PostIterator iter=f->newPostIterator(); !iter.atEnd(); ++iter) {
    Formula::Type type=(*iter).getType();
    if(bddmap.find(&(*iter))==bddmap.end()) {
      switch(type) {
      case Formula::fBinOp:
	switch(static_cast<const class BinOp &>(*iter).getOp()) {
	case BinOp::And:
	  result= bddmap[static_cast<const class BinOp &>(*iter).getLHS()] & 
	    bddmap[static_cast<const class BinOp &>(*iter).getRHS()];  
	  break;
	case BinOp::Or:
	  result= bddmap[static_cast<const class BinOp &>(*iter).getLHS()] | 
	    bddmap[static_cast<const class BinOp &>(*iter).getRHS()];
	  break;
	case BinOp::Impl:
	  result= bddmap[static_cast<const class BinOp &>(*iter).getLHS()] >> 
	    bddmap[static_cast<const class BinOp &>(*iter).getRHS()];
	  break;
	case BinOp::Equiv:
	  result= bdd_biimp(bddmap[static_cast<const class BinOp &>(*iter).getLHS()], 
	    bddmap[static_cast<const class BinOp &>(*iter).getRHS()]);
	  break;
	}
	bddmap.insert(FormulaBddMap::value_type(&(*iter), result));
	break;
      case Formula::fAtom: {
	const class Atom &tratt=static_cast<const Atom &>(*iter);
	if(mapping.find(&(*iter))!=mapping.end()) {
	  result=bdd_ithvar((*mapping.find(&(*iter))).second);
	}
	else {
	  result=!bdd_ithvar((*mapping.find(&Atom(tratt.getId(), !tratt.getSign()))).second);
	}
	bddmap.insert(FormulaBddMap::value_type(&(*iter), result));
	break;
      }
      case Formula::fConst:
	result=static_cast<const class Const &>(*iter).getVal() ? bdd_true() : bdd_false(); 
	bddmap.insert(FormulaBddMap::value_type(&(*iter), result));
	break;
      case Formula::fNot:
	result = !bddmap[static_cast<const class Not &>(*iter).getOperand()];
	bddmap.insert(FormulaBddMap::value_type(&(*iter), result));
	break;
      default:
	fputs("Parse error in gate expression\n", stderr);
	return bdd_false();
      }
    }
    else 
      result=bddmap[&(*iter)];
  }
  return result;
}


/**Iterate trough the formula and give bdd variable number to 
 * each atomic proposition. The variable numbers are assigned 
 * as for the safety automaton. The numbers are saved in the map.
 * @param formula
 * @param mapping Mapping from APs to bdd variable numbers 
 * @return The number of subformulas 
 */
inline static unsigned
apNumber(const class Formula &f, FormulaMap &mapping, AtomMap &idmap) 
{
  unsigned count=0;
  idmap.reserve(6);
  for(class Formula::Iterator i=f.newIterator(); !i.atEnd(); ++i) {
    if (idmap.capacity()<=count) idmap.reserve(2*idmap.capacity());
    if(mapping.find(&(*i))==mapping.end()) {
      if((*i).getType() == Formula::fAtom) {
	mapping.insert(FormulaMap::value_type(&(*i), count));
	idmap[count]=static_cast<class Atom &>(*i);
      }      
      else {
	mapping.insert(FormulaMap::value_type(&(*i), UINT_MAX));
      }
      count+=2;
    }
  }
  for(FormulaMap::iterator i=mapping.begin(); i!=mapping.end(); ) {
    if((*((*i).first)).getType() != Formula::fAtom) {
      mapping.erase(i);
      i=mapping.begin();
    }
    else 
      ++i;
  }

  return count;
}


typedef hash_map<unsigned, unsigned> NumberMap;
#define error(msg) fputs(msg, stderr); fclose(autfile); remove(outputFileName); delete aut; return 0;

  /**Pseudo-constructor for the class. The method invokes an external translator
   * and by sending the formula to stdout and reading the result of the invocation
   * from stdin. The formula is output the scheck format and the automaton read is
   * also assumed to be in scheck format. Note that this is not a member function.
   * @param translator path to the external translator
   * @param formula 
   * @return A (generalised) Bchi automaton corresponding to the formula
   */


ExplicitBuchi *
buchiRead(const class Formula &f, const char *translator,  
		    IntBddMap &labels, AtomMap &idmap) 
{

  char inputFileName[L_tmpnam+1];
  char outputFileName[L_tmpnam+1];

  if(tmpnam(inputFileName)==0 || tmpnam(outputFileName)==0) {
    fputs("Could not create temporary files.", stderr);
    return 0;
  }

  //assign var numbers to APs
  FormulaMap mapping;
  if(!bdd_isrunning()) bdd_init(500000, 50000);
  unsigned varNum=apNumber(f, mapping, idmap);
  if(static_cast<unsigned>(bdd_varnum()) < varNum) bdd_setvarnum(varNum);

  FILE *formulafile=fopen(inputFileName, "w+");
  if(formulafile==NULL) {
    fclose(formulafile);
    fputs("Could not open temporary files!\n", stderr); 
    return 0;    
  } 
  //print formula to file
  f.print(formulafile);
  fclose(formulafile);

  //create commanline
  char commandline[strlen(translator)+2*(L_tmpnam+1)+2]; commandline[0]='\0';
  strcat(commandline, translator); strcat(commandline, " "); strcat(commandline, inputFileName);
  strcat(commandline, " "); strcat(commandline, outputFileName);

  int errorcode=system(commandline);
  remove(inputFileName);

  if (errorcode != 0) {
    fprintf(stderr, "Invocation of the translator failed with error code %d.\n", errorcode);
  }

  FILE *autfile=fopen(outputFileName, "r");
  if(autfile==0) {
    fputs("Could not open temporary files!\n", stderr); 
    return 0;       
  }

  unsigned numStates=0; unsigned numSets=0;
  if(2!=fscanf(autfile, "%u%u", &numStates, &numSets)) {
    fputs("Parse error", stderr);
    fclose(autfile);
    remove(outputFileName);
    return 0;
  }
  if(!numStates) {
    fputs("Empty automaton\n", stderr);
    fclose(autfile); remove(outputFileName); 
    return 0;      
  }
   
  ExplicitBuchi *aut=new ExplicitBuchi(numStates, 1, numSets ? numSets : numSets+1);
  if(!numSets) for(unsigned i=numStates; i--;) aut->makeFinal(i,0);
  //read transiton information
  unsigned theInitial=UINT_MAX;
  NumberMap StateMap; //map: state -> new state nr
  NumberMap SetMap;  //map: set -> new set nr 
  BddMap ArcMap; //map: bdd -> index 
  int ch;
  for(unsigned i=numStates; i--;) {
    unsigned source, initial;
    source=initial=0;
    //int code=fscanf (autfile, "%u%u", &source, &initial);
    if (2 != fscanf (autfile, "%u%u", &source, &initial) || initial > 1) {
      error("Parse error\n");
    }
    if(StateMap.find(source)==StateMap.end()) {
      if(StateMap.size()==numStates) {
	fprintf(stderr, "Too many states seen by state %u\n", source);
	error(" ");
      }
      StateMap.insert(NumberMap::value_type(source, StateMap.size()));
    }

    if(initial) {
      if(theInitial != UINT_MAX) {
	error("Redefinition of the initial state.");
      }
      theInitial=StateMap[source];
    }

    //read set information
    for(;;) {
      unsigned setnum;
      while (isspace(ch=fgetc(autfile)));
      if(ch=='-') {
	if (1 != fscanf (autfile, "%u", &setnum) || 1 != setnum) {
          error("unexpected data after '-'\n");
        }
        break;
      }
      ungetc(ch, autfile);
      if(!isdigit(ch) || 1 != fscanf(autfile, "%u", &setnum) ) {
	error("Parse error");
      }
      if(SetMap.find(setnum) == SetMap.end()) { //the set has not been seen previously
	if(SetMap.size()==numSets) {
	  fprintf(stderr, "Too many sets seen by state %u and set %u.\n", source, setnum);
	  error(" ");
	}	
	aut->makeFinal(StateMap[source], SetMap.size());
	SetMap.insert(NumberMap::value_type(setnum, SetMap.size()));  
      }
      else {
	aut->makeFinal(StateMap[source], SetMap[setnum]);
      }
    }
    //read transition information
    for(;;) {
      unsigned dest;
      while (isspace(ch=fgetc(autfile)));
      if(ch=='-') {
	if (1 != fscanf (autfile, "%u", &dest) || 1 != dest) {
          fputs ("Unexpected data after '-'\n", stderr);
	  error(" ");
	  return 0;
        }
        break;
      }
      ungetc(ch, autfile);
      if(!isdigit(ch) || 1 != fscanf(autfile, "%u", &dest) ) {
	error("Parse error");
      }
      if(StateMap.find(dest) == StateMap.end()) { //the state has not been seen previously
	if(StateMap.size()==numStates) {
	  fprintf(stderr, "Too many states seen by state %u and transition to %u.\n", source, dest);
	  error("");
	}	
	StateMap.insert(NumberMap::value_type(dest, StateMap.size()));  
      }
      Formula *gate=parseGate(autfile);
      if(!gate) {error(" "); }
      class bdd label=createLabel(gate, mapping);
      if(ArcMap.find(label)==ArcMap.end()) { //previously unseen label
	labels.push_back(label);
	ArcMap.insert(BddMap::value_type(label , BddMap::data_type(ArcMap.size(), false)));
	if(aut->alphabetSize() < labels.size()) { 
	  aut->setASize(labels.size());
	}
      }
      aut->addTransition(StateMap[source], ArcMap[label].first, StateMap[dest]);
      gate->destroy();
    }    
  }

  while(isspace(ch=fgetc(autfile)));
  if(ch!=EOF) { 
    error("Extraneous non-whitespace data at end of input\n");
    return 0;
  }
  aut->makeInitial(theInitial);
  fclose(autfile);
  remove(outputFileName);
  return aut;
}




