//Abstract base class for finite automata represented explicitly
#ifndef EXPLICITBUCHI_H_
#define EXPLICITBUCHI_H_
#pragma interface
#include "BitVector.h"
#include "typedef.h"
#include <cstdio>


class ExplicitBuchi {
  
 public:
  /**Constructor of the class 
   * @param size The number of states in the automaton
   * @param aSize The size of the alphabet
   * @param setsize Number of acceptance sets
   */
  ExplicitBuchi(unsigned size, unsigned aSize, unsigned setsize);
  /**The destructor*/
  ~ExplicitBuchi(); 

  /**@return The number of states in the automaton*/
  unsigned size() const {return mySize;}
  /**@return The size of the alphabet*/
  unsigned alphabetSize() const {return myAlphabetSize;}  
   /**Get a successor state
   *@param source The source state
   *@param label The Label of the transition
   *@param num Index of the transition (for non-det automatons)
   *@return The destination state
   */
  unsigned dest(unsigned source, unsigned label, unsigned index) const {
    assert(source < mySize && label<myAlphabetSize && 
	   index<myTransRelSize[label*mySize + source]);
    return myTransRel[label*mySize + source][index];  
  }
  /**Add a transition 
   *@param source The source state
   *@param label The label of the transition
   *@param dest The destination state
   */
  void addTransition(unsigned source, unsigned label, unsigned dest);
  /**Delete a transition 
   *@param source The source state
   *@param label The label of the transition
   *@param dest The destination state
   */
  void delTransition(unsigned source, unsigned label, unsigned dest) {
    assert(source <= mySize && label<myAlphabetSize && dest<mySize);
    unsigned i=0;
    for(; dest!=myTransRel[label*mySize + source][i]; i++);
    myTransRel[label*mySize + source][i]
      =myTransRel[label*mySize + source][myTransRelSize[label*mySize + source]];
    myTransRel[label*mySize + source][myTransRelSize[label*mySize + source]--]=dest;
  }
  /**@return The number of transitions from a state with a given label*/ 
  unsigned numArcs(unsigned source, unsigned label) const {
    assert(source<mySize && label<myAlphabetSize);
    return myTransRelSize[label*mySize + source];   
  }
  /**Make a given state an initial state
   *@param state
   */
  void makeInitial(unsigned state) { 
    assert(state<mySize);
    myInitial.assign(state, true); 
  }
  /**Remove the initial state flag for a given state
   *@param state 
   */
  void delInitial(unsigned state) {
    assert(state<mySize);
    myInitial.assign(state, false); 
  }
  /**@return the number of an initial state*/
  unsigned initialState() const {
    unsigned i=0;
    for (; i<mySize; i++) {
      if (myInitial[i]) 
	break;
    }
    return i;
  }

  /**@return The number of acceptance sets*/
  unsigned numSets() const {return myNumAccSets;}

  /**Make a given state a final state
   *@param state The state 
   *@param set Which set the state should be added to ([0..myNumSets-1])
   */
  void makeFinal(unsigned state, unsigned set) {
    myFinal.assign(set*mySize+state , true);
  } 
  /**Remove a final state from a set
   *@param state The state
   *@param set The set to remove from
   */
  void delFinal(unsigned state, unsigned set) {
    myFinal.assign(set*mySize+state, false);
  }  
   /**@param state State to be examined
    * @return true if the state is a final state, otherwise false*/
  bool isFinal(unsigned state) const {
    assert(state<mySize);
    for (unsigned i=myNumAccSets; i--;) 
      if (myFinal[i*mySize+state]) 
	return true;
    return false;
  } 
  /**
   *@param state State to be examined  
   *@param set Set to be examined  
   *@return true if the state belongs to the given set
   */
  bool final(unsigned state, unsigned set) const {
    assert(state<mySize && set < myNumAccSets);
    return myFinal[set*mySize+state];
  }

  /**@return true if the state is an initial state*/
  bool isInitial(unsigned state) const {return myInitial[state]; }; 
  /**increase the size of the alphabet
   *@param size The new size of the alphabet
   */
  void setASize(unsigned size); 
   /**Print the automaton to a stream
   *@param output The stream where the output is directed
   *@param labels The labels of the arc in bdd form
   *@param idmap Names of the atomic proposition for the labels
   */
  void printAut(FILE *output, const IntBddMap &labels, const AtomMap &idmap) const;
   /**Print the automaton to a stream using numbers as labels
    *@param output The stream where the output is directed
    */
  void printAut(FILE *output) const;

 private:
  /**Number of states of the automaton*/
  unsigned mySize;
  /**Size of the alphabet*/
  unsigned myAlphabetSize;
  /**Number of acceptance sets*/
  unsigned myNumAccSets;
  /**The acceptance sets*/
  class BitVector myFinal;
  /**Array with the initial states*/  
  class BitVector myInitial;
  /**The transition relation*/
  unsigned *myTransRelSize;
  unsigned **myTransRel;
  /**Allocated alphabet size*/
  unsigned myAllocated;
};

ExplicitBuchi * buchiRead(const class Formula &f, const char *translator, 
			  IntBddMap &labels, AtomMap &idmap);

#endif //ABSAUTOMATA_H_

