#include "Pathologic.h"
#include "Formula.h"
#include "Buchi.h"
#include "ExplicitBuchi.h"
#include "NonDetAut.h"
#include "DetAut.h"
#include "BitVector.h"


Pathologic::Pathologic(const class Formula &formula, const class DetAut &detaut, const char *translator) 
  : myFormula(formula), myDetAut(detaut), myTranslator(translator) {}


Pathologic::~Pathologic() 
{
}


/**Check if state is in the map
 * @param visited The map to query
 * @param s The state to use in the query
 * @return 0 if s is not in visited, otherwise its depth number
 */
inline static unsigned
lookup(Pathologic::StateMap &visited, Pathologic::State &s)
{
  Pathologic::StateMap::const_iterator i=visited.find(s);
  return (i==visited.end()) ? 0 : (*i).second;

}

/**Initialise the complement
 * @param detaut Deterministic automaton to be complemented
 * @param labels Labels for the automaton
 * @return Complement of detaut interpreted as Buchi automaton
 */

inline static NonDetAut *
initComp(const DetAut &detaut, IntBddMap &labels) 
{
  unsigned i=0;
  for(  ; i<labels.size(); i++) {
    if(labels[i]==bdd_true()) 
      break;
  }
  if (i==labels.size()) labels.push_back(bdd_true());
  return detaut.buchiComplement(i);
}

/**Remove the potential change made to the labels
 * @param detaut Automaton owning the labels
 * @param labels map containing labels
 */
inline static void 
cleanUp(const DetAut &detaut, IntBddMap &labels)
{
  if (detaut.alphabetSize() != labels.size() ) { //true label was added 
    assert(labels.back() == bdd_true());
    labels.pop_back();
  }
  return;
}

bdd apBDD(const class Formula * f);
bdd labelBDD(const class Formula &f);


/**Make sure that the labels uses identical bdd variables
 *@param buchilabels The labels for the Buchi automaton
 *@param f The formula under inspection
 */
inline static void
initLabels(IntBddMap &buchilabels, const class Formula *f) 
{
  class bdd buchiAP=apBDD(f);
  class bdd compAP=labelBDD(*f);

  int *v1, *v2, n1, n2;
  bdd_scanset(buchiAP, v1, n1);
  bdd_scanset(compAP, v2, n2);
  assert(n1 == n2);
  
  bddPair *pair=bdd_newpair();
  bdd_setpairs(pair, v1, v2, n1);

  for(IntBddMap::iterator i=buchilabels.begin(); i != buchilabels.end(); ++i) {
    *i=bdd_replace(*i, pair);    
  }
  
  bdd_freepair(pair);
  free(v1);
  free(v2);
}


ExplicitBuchi* testAut(IntBddMap &labels, AtomMap &idmap);

/**Check if the negation of the given formula is pathologic
 *@param labels Labels in bdd form for the deterministic automaton
 *@return true if the formula is ok (not pathologic), otherwise false
 */
 
bool
Pathologic::pathologic(IntBddMap &labels) const
{ 
  /**Storage for dfs numbers*/
  StateMap visited; 
  /**Father stack */
  StateStack father;
  /**Dfs stack*/
  DepthStack dfsStack;
  /**Temporary storage for successors*/
  StateStack children;
  /**Current Depth first search number*/
  unsigned depth=1;
  /**Flag if state has a selfloop*/
  class BitVector selfloops;

  class NonDetAut *complement=initComp(myDetAut, labels);
  IntBddMap buchilabels;
  AtomMap idmap;
  class ExplicitBuchi *buchi= (myTranslator) ? buchiRead(myFormula, myTranslator, buchilabels, idmap) : 0;
  class Buchi *symbolicBuchi=0;
  if(!buchi) {
    symbolicBuchi=Buchi::create(myFormula);
    buchi=symbolicBuchi->toExplicit(buchilabels, idmap);
  } 
  if (!myTranslator) initLabels(buchilabels, &myFormula);
  /**Current state of the product automaton*/
  class State s(buchi->initialState(), complement->initialState());
  while (true) {
    // compute a strongly connected component
    while (true) {
      std::pair<StateMap::iterator,bool> p =
        visited.insert (StateMap::value_type (s, depth));
      if (!p.second) break;
      father.push_front (s);
      unsigned enabled=0;
      for (unsigned i=complement->alphabetSize(); i--;) {
	for(unsigned j=buchi->alphabetSize(); j--; ) {
	  if ((labels[i] & buchilabels[j])!=bdd_false()) {
	    //fputs("Comp", stderr); bdd_fprintset(stderr, labels[i]);
	    //fputs("Buchi", stderr);bdd_fprintset(stderr, buchilabels[j]);
	    //fputs("\n", stderr);
	    for (unsigned k=complement->numArcs(s.complement, i); k--; ) {
	      for (unsigned l=buchi->numArcs(s.buchi, j); l--; ) {	
		//fputs("And", stdout);bdd_printset(labels[i]&buchilabels[j]);
		children.push_front (State(buchi->dest(s.buchi, j, l), 
					   complement->dest(s.complement, i, k)));
		enabled++;
		if (s.buchi == buchi->dest(s.buchi, j, l) 
		    && s.complement == complement->dest(s.complement, i, k)) 
		  selfloops.extend (depth, true);
	      }
	    }
	  }
	}
      }
      dfsStack.push_front (Depth (s, depth++, enabled));
      if (!enabled) break;     
      s = children.front ();
      children.pop_front ();
    }
    unsigned minDepth = visited[s];
    
    // backtrack
    for (;;) {
      if (dfsStack.empty ()) {	
	delete complement;
	delete buchi;
	cleanUp(myDetAut, labels);
	if(myTranslator) delete symbolicBuchi;
	return true;
      }
      class Depth& tarjan = dfsStack.front ();
      unsigned sDepth = tarjan.dfs;     
      if (minDepth && minDepth < sDepth)
	tarjan.dfs = sDepth = minDepth;     
      if (tarjan.numSucc-- > 1) {
	assert (!children.empty ());
	s = children.front ();
	children.pop_front ();
	break;
      }
      s = tarjan.state;
      dfsStack.pop_front ();
      
      if (sDepth == ::lookup (visited, s)) { //component found
	unsigned maxDepth = visited[father.front ()];
	assert (maxDepth >= sDepth);
	selfloops.setSize (maxDepth + 1);
	class BitVector visitedSets (buchi->numSets ()+1);
	for (depth = maxDepth + 1; depth > sDepth; depth--) { 
	  const State &state=father.front();
	  visited[state]=0;
	  for (unsigned set=buchi->numSets(); set--; ) { //gen buchi sets are from 1..numBuchiSets
	    if (buchi->final(state.buchi, set)) visitedSets.assign (set+1, true);	    
	  }
	  if (complement->isFinal(state.complement)) visitedSets.assign (0, true); //complement set is 0	    
	  father.pop_front ();	  	  		  
	}
	if ((maxDepth>sDepth || selfloops[maxDepth]) && visitedSets.allSet ()) {	  
	  delete complement;
	  delete buchi;	    
	  cleanUp(myDetAut, labels);
	  if(myTranslator) delete symbolicBuchi;
	  return false;	  
	}
	selfloops.truncate (depth = sDepth);
      }      
      minDepth = tarjan.dfs;
    }
  }  
  return true;
} 
