/**Container class for methods related to analysing the
   pathologicness of an LTL formula*/
#include "typedef.h"
#include <slist>

//Forward decs
class Formula;
class Buchi;
class DetAut;
class NonDetAut;

class Pathologic {
  
 public:
  /**Constructor of the class
   *@param formula The negation of the property, i.e. the formula given
   * to scheck
   *@param detaut Deterministic finite automaton representing the formula
   *@param translator Commandline to execute external translator
   */
  Pathologic(const class Formula &formula, const class DetAut &detaut, const char *translator);  
  
  /**The destructor*/
  ~Pathologic();

  /**Determine if the given formula is pathologic
   *@param labels Bdd labels of the deterministic automaton
   *@return true if the formula is NOT pathologic, otherwise false
   */
  bool pathologic (IntBddMap &labels) const;
  
  /**Class for representing product state*/
  class State {
  public:   
    /**The constructor 
     *@param b State number for the Buchi automaton
     *@param c State number for the buchi complemented dfa
     */
    State(unsigned b, unsigned c) : buchi(b), complement(c) {}

    /**The copy constructor
     *@param orig The original to be copied
     */
    State(const class State &orig) : buchi(orig.buchi), 
      complement(orig.complement) {} 
    
    /**Equality comparison
     *@param other State to be compared with
     *@return true if the States are equal
     */
    bool operator==(const class State &other) const {
      return buchi==other.buchi && complement==other.complement;    
    }
    
    /**The assignment operator
     *@param rhs The right-hand side operand 
     *@return A reference to this object
     */
    State & operator=(const class State rhs) {
      buchi=rhs.buchi;
      complement=rhs.complement;      
      return *this;
    }    
 
    /**@return the hash value of the state*/
    size_t operator() () const {
      return hasher (buchi) ^ hasher (complement);
    }

    
  public:
    unsigned buchi;
    unsigned complement;    

  private:
    /** Integer hash function by Robert Jenkins (bob_jenkins@compuserve.com)
     * <URL:http://ourworld.compuserve.com/homepages/bob_jenkins/blockcip.htm>
     * @param key       key to be hashed
     * @return          hash value of key
     */
    size_t hasher (size_t key) const {
      key += key << 12;
      key ^= key >> 22;
      key += key << 4;
      key ^= key >> 9;
      key += key << 10;
      key ^= key >> 2;
      key += key << 7;
      key ^= key >> 12;
      return key;
    }
  };

  /**Class for the state of the non-recursive dfs*/
  class Depth {
  public:
    /**The constructor*/
    Depth(class State &s, unsigned depth, unsigned succs) :
      state(s), dfs(depth), numSucc(succs) {}

    /**Copy constructor*/
    Depth(const class Depth &depth) : state(depth.state),
      dfs(depth.dfs), numSucc(depth.numSucc) {}

    /**Equality comparison*/
    bool operator==(const class Depth &rhs) const {
      return state==rhs.state && dfs==rhs.dfs && numSucc==rhs.numSucc;
    }

    /**Assignment operator*/
    class Depth & operator=(const class Depth &rhs) {
      state=rhs.state;
      dfs=rhs.dfs;
      numSucc=rhs.numSucc;
      return *this;
    } 
    /**State of the search*/
    class State state;
    /**Current dfs number of the state*/
    unsigned dfs;
    /**Number of unexplored successors*/
    unsigned numSucc;
  };
  
 public:
  struct statehash
  {
    /** Calculate the hash value of a state
     * @param state     state to be hashed
     * @return          the hash value
     */
    size_t operator() (const class State& state) const {
      return state ();
    }
  };

  typedef std::hash_map<class State, unsigned, statehash> StateMap;
  typedef std::slist<State> StateStack;
  typedef std::slist<Depth> DepthStack;

 private:
  /**Formula under consideration*/
  const class Formula &myFormula;
  /**Simple Buhci automaton representing the complemented automaton*/
  const class DetAut &myDetAut;
  /**Command line for executing external translator*/
  const char *myTranslator;
};
